import dataclasses
from typing import Literal, Optional

import numpy as np
import scipy.interpolate
import scipy.signal

from .data import Signal


def cubic_spline_interpolation(signal: Signal, resample_rate: int) -> Signal:
    if resample_rate % signal.sample_rate != 0:
        raise RuntimeError(
            f"Cannot resample from {signal.sample_rate = }Hz to {resample_rate = }Hz: "
            f"{resample_rate % signal.sample_rate = } must be zero."
        )

    sample_ratio = resample_rate / signal.sample_rate
    result_size = len(signal.data) * sample_ratio
    b_spline = scipy.interpolate.make_interp_spline(
        np.arange(0, result_size, sample_ratio), signal.data
    )
    return Signal(data=b_spline(np.arange(0, result_size)), sample_rate=resample_rate)


def butterworth_filter(
    signal: Signal,
    cutoff_freq: float,
    filter_type: Literal["highpass", "lowpass"],
) -> Signal:
    nyquist_freq = signal.sample_rate / 2
    sos = scipy.signal.butter(
        N=5, Wn=(cutoff_freq / nyquist_freq), btype=filter_type, output="sos"
    )
    return dataclasses.replace(signal, data=scipy.signal.sosfiltfilt(sos, signal.data))


def sg_filter(signal: Signal, sg_settings: tuple[int, int]) -> Signal:
    poly_order, smoothing_window_ms = sg_settings
    smoothing_window = (smoothing_window_ms / 1000) * signal.sample_rate
    smoothing_window = round(smoothing_window)

    # smoothing_window must be odd
    if smoothing_window % 2 == 0:
        smoothing_window += 1

    return dataclasses.replace(
        signal, data=scipy.signal.savgol_filter(signal.data, smoothing_window, poly_order)
    )


def preprocess(
    signal: Signal,
    resample_rate: Optional[int] = 1000,
    highpass_cutoff: Optional[float] = 0.5,
    lowpass_cutoff: Optional[float] = None,
    sg_settings: Optional[tuple[int, int]] = (3, 100),
) -> Signal:
    """Prepares cardiac data for analysis using global functions.

    Applies in order:
    cubic spline interpolation,
    highpass and lowpass Butterworth filters
    and Savitzky-Golay smoothing.

    Parameters set to None imply that aspect of the pipeline will not be applied.
    For example, the default value for `lowpass_cutoff` is None,
    which implies that by default the lowpass filter will not be applied.

    Parameters
    ----------
    signal : Signal
        Cardiac signal to be processed.
    resample_rate : int, default: 1000
        If greater than `signal.sample_rate`,
        will be used as the target sample rate (hertz) for cubic spline interpolation.
        Must be divisible by `signal.sample_rate`.
    highpass_cutoff : float, default: 0.5
        Butterworth highpass filter cutoff frequency in hertz.
    lowpass_cutoff : float, optional
        Butterworth lowpass filter cutoff frequency in Hertz, filter is off by default.
    sg_settings : (int, int), default: (3, 100)
        Savitzky-Golay smoothing parameters,
        where the first element of the tuple is the polynomial order
        and the second is the window size in milliseconds.

    Returns
    -------
    array_like
        Preprocessed signal
    """
    nans = np.isnan(signal.data)
    if np.any(nans):
        raise RuntimeError(
            "Cannot preprocess data containing NaN values. "
            f"First NaN found at index {nans.nonzero()[0][0]}."
        )

    if resample_rate is not None and resample_rate > signal.sample_rate:
        result = cubic_spline_interpolation(signal, resample_rate)
    else:
        result = signal

    if highpass_cutoff is not None:
        result = butterworth_filter(result, highpass_cutoff, "highpass")

    if lowpass_cutoff is not None:
        result = butterworth_filter(result, lowpass_cutoff, "lowpass")

    if sg_settings:
        result = sg_filter(result, sg_settings)

    return result



from typing import Union

import numpy as np
import pandas as pd
import scipy.interpolate
import scipy.signal
import scipy.stats
import sklearn.cluster
import sklearn.preprocessing

from .data import OutlierDetectionSettings, Signal

DATA_COLUMNS = ["BPM", "RMSSD", "SDNN", "SDSD", "pNN20", "pNN50", "HF"]
DATAFRAME_COLUMNS = ["Time", *DATA_COLUMNS, "Outlier", "Window"]


def analyze(
    signal: Signal,
    window_width: int = 10,
    window_overlap: int = 0,
    ecg_prt_clustering: bool = False,
    amplitude_threshold: int = 50,
    distance_threshold: int = 250,
    n_required_peaks: int = 3,
    outlier_detection_settings: Union[str, OutlierDetectionSettings] = "moderate",
) -> pd.DataFrame:
    """Analyzes cardiac data.

    Extracts BPM, RMSSD and SDNN from `input_data`.

    Parameters
    ----------
    signal : Signal
        Cardiac signal to be analyzed.
    window_width : int, default: 10
        Width of the sliding window in seconds.
    window_overlap: int, default: 0
        Amount of overlap between windows in seconds.
        Accepts negative values, interpreted as space between windows.
    ecg_prt_clustering: bool, default: False
        Use k-means clustering to detect P, R and T waves in the data.
        Useful for atypical morphologies (e.g. T amplitude > R amplitude).
        If enabled, `amplitude_threshold` and `distance_threshold` will be ignored.
    amplitude_threshold: int, default: 50
        Minimum signal amplitude for a peak to be registered.
        For PPG data, the recommended value is 30.
    distance_threshold: int, default: 250
        Minimum time in milliseconds since last peak for a new peak to be registered.
    n_required_peaks: int, default: 3
        Minimum number of peaks in a window required to record analysis for that window.
        Values less than three are invalid.
    outlier_detection_settings: str or OutlierDetectionSettings, default: "moderate"
        Settings for the Outlier detection algorithm.
        Accepts either an `OutlierDetectionSettings` object, or a string specifying a method.
        Refer to :class:`OutlierDetectionSettings` for details.

    Returns
    -------
    Dataframe containing Extracted heart data.
    """
    # Validate arguments
    outlier_detection_settings = (
        OutlierDetectionSettings.from_method(outlier_detection_settings)
        if isinstance(outlier_detection_settings, str)
        else outlier_detection_settings
    )

    if n_required_peaks < 3:
        raise ValueError("Parameter 'n_required_peaks' must be greater than three.")

    # Peak detection settings
    if ecg_prt_clustering:
        distance = 1
        prominence = 5
    else:
        distance = int((distance_threshold / 1000) * signal.sample_rate)
        prominence = amplitude_threshold

    # Windowing function
    results = []
    for sample_start in range(
        0, len(signal.data), (window_width - window_overlap) * signal.sample_rate
    ):
        timestamp = sample_start / signal.sample_rate

        segment = signal.data[sample_start : sample_start + (window_width * signal.sample_rate)]
        normalized = sklearn.preprocessing.minmax_scale(segment, (0, 100))
        peaks, properties = peak_detection(normalized, distance, prominence, ecg_prt_clustering)
        window_data = (normalized, peaks, properties)

        ibi = np.diff(peaks) * 1000 / signal.sample_rate
        sd = np.diff(ibi)

        if len(peaks) <= n_required_peaks:
            results.append([timestamp, *[np.nan] * len(DATA_COLUMNS), True, window_data])
        else:
            # Time-domain metrics
            bpm = ((len(peaks) - 1) / ((peaks[-1] - peaks[0]) / signal.sample_rate)) * 60
            rmssd = np.sqrt(np.mean(np.square(sd)))
            sdnn = np.std(ibi)
            sdsd = np.std(sd)  # Standard deviation of successive differences
            p_nn20 = np.sum(sd > 20) / len(sd)  # Proportion of successive differences > 20ms
            p_nn50 = np.sum(sd > 50) / len(sd)  # Proportion of successive differences > 50ms

            # Frequency-domain metrics
            hf = frequency_domain(x=ibi, sfreq=signal.sample_rate)

            is_outlier = outlier_detection(
                peaks,
                properties,
                ibi,
                signal.sample_rate,
                window_width,
                bpm,
                rmssd,
                outlier_detection_settings,
            )

            results.append(
                [timestamp, bpm, rmssd, sdnn, sdsd, p_nn20, p_nn50, hf, is_outlier, window_data]
            )

    return pd.DataFrame(
        results,
        columns=DATAFRAME_COLUMNS,
    )


def peak_detection(
    segment: np.ndarray, distance: int, prominence: int, use_clustering: bool
) -> tuple[np.ndarray, dict]:
    """Returns the indexes of detected peaks and associated properties."""
    peaks, properties = scipy.signal.find_peaks(
        segment, distance=distance, prominence=prominence, height=0, width=0
    )

    # Attempt to determine correct peaks by distinguishing the R wave from P and T waves
    if len(peaks) >= 3 and use_clustering:
        k_means = sklearn.cluster.KMeans(n_clusters=3).fit(
            np.column_stack(
                (properties["widths"], properties["peak_heights"], properties["prominences"])
            )
        )

        # Use width centroids to determine correct wave (least width, most prominence)
        # If the two lowest values are too close (< 5), use prominence to distinguish them
        width_cen = k_means.cluster_centers_[:, 0]
        labels_sort_width = np.argsort(width_cen)
        if width_cen[labels_sort_width[1]] - width_cen[labels_sort_width[0]] < 5:
            # Label of maximum prominence for lowest two widths
            prom_cen = k_means.cluster_centers_[:, 2]
            wave_label = np.argsort(prom_cen[labels_sort_width[:2]])[1]
        else:
            wave_label = labels_sort_width[0]

        is_wave_peak = k_means.labels_ == wave_label

        wave_peaks = peaks[is_wave_peak]
        wave_props = {k: v[is_wave_peak] for k, v in properties.items()}
    else:
        wave_peaks = peaks
        wave_props = properties

    # @PeterKirk does this need to be > 3 or >= 3?
    # Also, should this potentially be done before clustering?
    if len(wave_peaks) > 3:
        # Approximate prominences at edges of window
        base_height = segment[wave_peaks] - wave_props["prominences"]
        wave_props["prominences"][0] = wave_props["peak_heights"][0] - base_height[1]
        wave_props["prominences"][-1] = wave_props["peak_heights"][-1] - base_height[-2]

    return wave_peaks, wave_props


def frequency_domain(x, sfreq: int = 5):
    """This function and docstring was modified from Systole
    (https://github.com/embodied-computation-group/systole)
    Extracts the frequency domain features of heart rate variability.
    Parameters
    ----------
    x : np.ndarray or list
        Interval time-series (R-R, beat-to-beat...), in miliseconds.
    sfreq : int
        The sampling frequency (Hz).
    Returns
    -------
    stats : :py:class:`pandas.DataFrame`
        Frequency domain summary statistics.
        * ``'power_hf_per'`` : High frequency power (%).
    Notes
    -----
    The dataframe containing the summary statistics is returned in the long
    format to facilitate the creation of group summary data frame that can
    easily be transferred to other plotting or statistics library. You can
    easily convert it into a wide format for a subject-level inline report
    using the py:pandas.pivot_table() function:
    >>> pd.pivot_table(stats, values='Values', columns='Metric')
    """
    if len(x) < 4:  # RapidHRV edit: Can't run with less than 4 IBIs
        return np.nan

    # Interpolate R-R interval
    time = np.cumsum(x)
    f = scipy.interpolate.interp1d(time, x, kind="cubic")
    new_time = np.arange(time[0], time[-1], 1000 / sfreq)  # sfreq = 5 Hz
    x = f(new_time)

    # Define window length
    nperseg = 256 * sfreq
    if nperseg > len(x):
        nperseg = len(x)

    # Compute Power Spectral Density
    freq, psd = scipy.signal.welch(x=x, fs=sfreq, nperseg=nperseg, nfft=nperseg)
    psd = psd / 1000000
    fbands = {"hf": ("High frequency", (0.15, 0.4), "r")}

    # Extract HRV parameters
    ########################
    stats = pd.DataFrame([])
    band = "hf"

    this_psd = psd[(freq >= fbands[band][1][0]) & (freq < fbands[band][1][1])]
    this_freq = freq[(freq >= fbands[band][1][0]) & (freq < fbands[band][1][1])]

    if (len(this_psd) == 0) | (len(this_psd) == 0):  # RapidHRV edit: if no power
        return np.nan

    # Peaks (Hz)
    peak = round(this_freq[np.argmax(this_psd)], 4)
    stats = stats.append({"Values": peak, "Metric": band + "_peak"}, ignore_index=True)

    # Power (ms**2)
    power = np.trapz(x=this_freq, y=this_psd) * 1000000
    stats = stats.append({"Values": power, "Metric": band + "_power"}, ignore_index=True)

    hf = stats.Values[stats.Metric == "hf_power"].values[0]

    return hf


def outlier_detection(
    peaks: np.ndarray,
    peak_properties: dict,
    ibi: np.ndarray,
    sample_rate: int,
    window_width: int,
    bpm: float,
    rmssd: float,
    settings: OutlierDetectionSettings,
) -> bool:
    bpm_in_range = settings.bpm_range[0] < bpm < settings.bpm_range[1]
    rmssd_in_range = settings.rmssd_range[0] < rmssd < settings.rmssd_range[1]
    if not (bpm_in_range and rmssd_in_range):
        return True

    max_peak_distance = (peaks[-1] - peaks[0]) / sample_rate
    if max_peak_distance < (window_width * settings.min_total_peak_distance):
        return True

    def mad_outlier_detection(x: np.ndarray, threshold: float) -> np.ndarray:
        x = x - np.mean(x)
        mad = scipy.stats.median_abs_deviation(x) * threshold
        return (x > mad) | (x < -mad)

    prominence_outliers = mad_outlier_detection(
        peak_properties["prominences"], settings.mad_threshold
    )
    if np.any(prominence_outliers):
        return True

    height_outliers = mad_outlier_detection(
        peak_properties["peak_heights"], settings.mad_threshold
    )
    if np.any(height_outliers):
        return True

    ibi_outliers = mad_outlier_detection(ibi, settings.ibi_mad_threshold)
    if np.any(ibi_outliers):
        return True

    return False



from .analysis import analyze
from .data import OutlierDetectionSettings, Signal, get_example_data
from .preprocessing import preprocess
from .visualization import visualize

__all__ = (
    "analyze",
    "OutlierDetectionSettings",
    "Signal",
    "get_example_data",
    "preprocess",
    "visualize",
)



from __future__ import annotations

import dataclasses

import h5py
import numpy as np
import pandas as pd


@dataclasses.dataclass
class OutlierDetectionSettings:
    """Settings for outlier detection.

    Attributes
    ----------
    bpm_range:
        Range of acceptable bpm values.
    rmssd_range:
        Range of acceptable rmssd values.
    mad_threshold:
        Threshold for peak heights and prominences to register as outliers.
        (in median absolute deviation units)
    ibi_mad_threshold:
        Threshold for peak intervals to register as an outliers.
        (in median absolute deviation units)
    min_total_peak_distance:
        Acceptable ratio between total width to distance between first and last peaks in window.
    """

    bpm_range: tuple[int, int]
    rmssd_range: tuple[int, int]
    mad_threshold: int
    ibi_mad_threshold: int
    min_total_peak_distance: float = 0.5

    @classmethod
    def from_method(cls, method: str) -> OutlierDetectionSettings:
        """Generate settings from method name.

        Method names are: "liberal", "moderate", "conservative".
        "conservative" is the most stringent, "liberal" is the least and "moderate" is in-between.
        """
        if method == "liberal":
            return OutlierDetectionSettings(
                bpm_range=(20, 200), rmssd_range=(0, 300), mad_threshold=7, ibi_mad_threshold=7
            )
        elif method == "moderate":
            return OutlierDetectionSettings(
                bpm_range=(30, 190), rmssd_range=(5, 262), mad_threshold=5, ibi_mad_threshold=5
            )
        elif method == "conservative":
            return OutlierDetectionSettings(
                bpm_range=(40, 180), rmssd_range=(10, 200), mad_threshold=4, ibi_mad_threshold=4
            )
        else:
            raise RuntimeError(f"Invalid outlier detection method: {method}.")


@dataclasses.dataclass
class Signal:
    """Raw signal with associated metadata.

    Attributes
    ----------
    data:
        Raw signal data.
    sample_rate:
        Signal rate in Hertz of raw signal.
    """

    data: np.ndarray
    sample_rate: int

    def __post_init__(self):
        self.data = self.data if isinstance(self.data, np.ndarray) else np.array(self.data)

    def save(self, filename: str) -> None:
        """Save as filename.hdf5"""
        with h5py.File(filename, "w") as f:
            f["data"] = self.data
            f.attrs["sample_rate"] = self.sample_rate

    @classmethod
    def load(cls, filename: str) -> Signal:
        """Load from filename.hdf5"""
        with h5py.File(filename, "r") as f:
            return cls(
                data=f["data"],
                sample_rate=int(f.attrs["sample_rate"]),
            )

    @classmethod
    def from_csv(cls, filename: str, sample_rate: int):
        data = pd.read_csv(filename).to_numpy()[0]
        return cls(data=data, sample_rate=sample_rate)

    @classmethod
    def from_txt(cls, filename: str, sample_rate: int):
        data = np.loadtxt(filename)
        return cls(data=data, sample_rate=sample_rate)


def get_example_data() -> Signal:
    """Function to get example data from `OSF <https://osf.io>`

    Returns
    -------
    array_like
        example data
    """
    return Signal.from_csv("https://osf.io/wqnjh/download", sample_rate=20)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['rapidhrv']

package_data = \
{'': ['*']}

install_requires = \
['dash>=2.0.0,<3.0.0',
 'h5py>=3.3.0,<4.0.0',
 'numpy>=1.21.0,<2.0.0',
 'pandas>=1.3.0,<2.0.0',
 'scikit-learn>=0.24.2,<0.25.0',
 'scipy>=1.7.0,<2.0.0']

extras_require = \
{'notebooks': ['jupyter>=1.0.0,<2.0.0', 'matplotlib>=3.4.2,<4.0.0']}

setup_kwargs = {
    'name': 'rapidhrv',
    'version': '0.2.3',
    'description': 'A package for preprocessing, analyzing and visualizing cardiac data',
    'long_description': '# RapidHRV\n\nRapidHRV is a data processing pipeline for the analysis and visualization of cardiac data.\n\nPlease provide credit where appropriate:\n\nKirk, P. A., Davidson Bryan, A., Garfinkel, S., & Robinson, O. J. (2021).\n_RapidHRV: An open-source toolbox for extracting heart rate and heart rate variability_\n([PsyArXiv](https://doi.org/10.31234/osf.io/3ewgz))\n\nThis library is distributed under an \n[MIT License](https://raw.githubusercontent.com/peterakirk/RapidHRV/main/LICENSE)\n\n## Installation\n\n```shell\npip install rapidhrv\n```\n\n## Usage\n\nGiven a numpy array, or something convertable to it (such as a list),\n`rapidhrv.preprocess` can generate input suitable for analysis with\n`rapidhrv.analyze`, which will return a pandas dataframe containing HRV data.\n\n```python\nimport numpy as np\nimport rapidhrv as rhv\n\nmy_data = np.load("my_data.npy")  # Load data\ndata = rhv.Signal(my_data, sample_rate=50)  # Convert to rhv Signal class\npreprocessed = rhv.preprocess(data)  # Preprocess: may interpolate data, check the docstring on `rapidhrv.preprocess`\nresult = rhv.analyze(preprocessed)  # Analyze signal\n```\n\n## Documentation\n\nPlease see the included [tutorial notebook](https://github.com/peterakirk/RapidHRV/blob/main/resources/tutorial.ipynb).\n\n## Development\n\nIn order to get a working development environment,\nplease install [Poetry](https://python-poetry.org/) for your platform,\nand run `poetry install` to generate a virtual environment.\n\nIf you plan on making any changes to the included notebooks,\nplease run `nbstripout --install` from within the poetry venv before committing any changes.\n\nTo run said notebooks from the environment provided by poetry,\ninstall the required dependencies with `poetry install --extras notebooks`.\n\n',
    'author': 'Peter Kirk',
    'author_email': 'p.kirk@ucl.ac.uk',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.9,<3.10',
}


setup(**setup_kwargs)



import dash
import pandas as pd
import plotly.graph_objects as go
from dash import dcc, html
from dash.dependencies import Input, Output

import rapidhrv as rhv


def results_graph(non_outliers, outliers, selected_column):
    fig = go.Figure(
        [
            go.Scatter(
                x=non_outliers["Time"],
                y=non_outliers[selected_column],
                name=selected_column,
                mode="lines+markers",
            ),
            go.Scatter(
                x=outliers["Time"],
                y=outliers[selected_column],
                name="Outliers",
                mode="markers",
            ),
        ]
    )

    fig.update_layout(template="plotly_white", clickmode="event+select")
    fig.update_traces(marker_size=10)

    return fig


def window_graph(window_data):
    signal, peaks, properties = window_data
    fig = go.Figure(
        [
            go.Scatter(y=signal),
            go.Scatter(
                x=peaks,
                y=properties["peak_heights"],
                mode="markers",
                marker=dict(line_color="darkorange", symbol="x-thin", size=24, line_width=1.5),
            ),
        ],
    )
    fig.update_layout(showlegend=False, template="plotly_white")

    return fig


def visualize(analyzed: pd.DataFrame, debug=False):
    app = dash.Dash()

    non_outlier_data = analyzed.loc[~analyzed["Outlier"]]
    outlier_data = analyzed.loc[analyzed["Outlier"]]

    selected_column = "BPM"
    results = results_graph(non_outlier_data, outlier_data, selected_column)

    app.layout = html.Div(
        [
            dcc.Dropdown(
                id="column-dropdown",
                options=[{"label": col, "value": col} for col in rhv.analysis.DATA_COLUMNS],
                value=selected_column,
                clearable=False,
            ),
            dcc.Graph(id="results-graph", figure=results),
            html.Div(id="window-container"),
        ]
    )

    @app.callback(Output("results-graph", "figure"), Input("column-dropdown", "value"))
    def update_results_graph(column):
        return results_graph(non_outlier_data, outlier_data, column)

    @app.callback(Output("window-container", "children"), Input("results-graph", "clickData"))
    def update_window_graph(click_data):
        if click_data is None:
            return []

        selected_point = click_data["points"][0]

        if selected_point["curveNumber"] == 0:
            window_data = non_outlier_data.iloc[selected_point["pointNumber"]]["Window"]
        elif selected_point["curveNumber"] == 1:
            window_data = outlier_data.iloc[selected_point["pointNumber"]]["Window"]

        return [dcc.Graph(figure=window_graph(window_data))]

    app.run_server(debug=debug, dev_tools_silence_routes_logging=True)


if __name__ == "__main__":
    signal = rhv.data.get_example_data()
    preprocessed = rhv.preprocess(signal)
    analyzed = rhv.analyze(preprocessed)
    visualize(analyzed, debug=True)
